/**
 * \file scoped.h scoped resource.
 *
 * Written by Kevlin Henney with small modifications by Martin Moene.
 *
 * This code is presented in the following articles.
 *
 * - Kevlin Henney. Making an Exception, Effective C++ column in
 *   Application Development Advisor 5(4), May 2001
 *   http://www.two-sdg.demon.co.uk/curbralan/papers/MakingAnException.pdf
 *
 * - Kevlin Henney. One Careful Owner, Effective C++ column in
 *   Application Development Advisor 5(5), June 2001.
 *   http://www.two-sdg.demon.co.uk/curbralan/papers/OneCarefulOwner.pdf
 */

#ifndef _scoped_h_included
#define _scoped_h_included

/**
 * scoped<> really should be in the std namespace.
 */
namespace nonstd {

/**
 * implementation details.
 */
namespace detail {

/**
 * inherit from this class to prevent copying.
 */
class noncopyable
{
protected:
   /**
    * default constructible.
    */
    noncopyable() {}

private:
   /**
    * non copy-constructible.
    */
    noncopyable( noncopyable const& );

   /**
    * non copy-assingable.
    */
    noncopyable& operator= ( noncopyable const& );
};

} // namespace detail

/**
 * dereferenced element type trait.
 */
template< typename value_type >
struct dereferenced_trait
{
   typedef void type;           ///< the dereferenced type.
};

/**
 * dereferenced element type trait, pointer partial specialization.
 */
template< typename element_type >
struct dereferenced_trait< element_type * >
{
   typedef element_type& type;  ///< the dereferenced type.
};

/**
 * dereferenced element type trait, void pointer specialization.
 */
template<>
struct dereferenced_trait< void * >
{
   typedef void type;           ///< the dereferenced type.
};

/**
 * standard delete functor.
 */
struct deleter
{
   /**
    * delete given object.
    */
   template< typename deleted_type >
   void operator()( deleted_type to_delete ) const
   {
      delete to_delete;
   }
};

/**
 * scoped resource class.
 */
template
   < typename resource_type
   , typename destructor_type = deleter
   >
class scoped : detail::noncopyable
{
public:
//    /// when included, doxygen 1.5.1-p1 does not show this section.
//    /// \name Construction
//    /// @{

   /**
    * default constructor.
    */
   scoped()
   : to_destruct( false )
   {
   }

   /**
    * constructor.
    */
   explicit scoped( resource_type resourced, destructor_type on_destruct = destructor_type() )
   : to_destruct( true )
   , resource( resourced )
   , destruct( on_destruct )
   {
   }

   /**
    * destructor.
    */
   ~scoped()
   {
      if ( to_destruct )
      {
         destruct( resource );
      }
   }

   /// @}
   /// \name Conversion
   /// @{

   /**
    * return resource.
    */
   operator resource_type() const
   {
      return resource;
   }

   /**
    * return resource.
    */
   resource_type operator->() const
   {
      return resource;
   }

   /**
    * dereference resource.
    */
   typename dereferenced_trait<resource_type>::type operator*() const
   {
      return *resource;
   }

#if defined HAVE_SCOPED_ADDRESSOF_CONVERSION
   /**
    * return address of resource; see detailed description below.
    *
    * This operator is only available if the \c HAVE_SCOPED_ADDRESSOF_CONVERSION
    * macro is defined.
    *
    * It is an extension made by Richard Howells to better support Windows programmers.
    * He added support for operator& because, however ugly, it is the classic way
    * for a Windows API function to return a \c HANDLE.
    *
    * See \ref howells_raii "[Howells RAII]" for more information.
    */
   resource_type * operator&()
   {
      return &resource;
   }
#endif

   /// @}
   /// \name Transfer
   /// @{

   /**
    * scoped resource proxy to transfer resource ownership without leaking if
    * the transfered resource is not being used.
    */
   class result
   {
   public:
      /**
       * constructor.
       */
      result( resource_type resourced, destructor_type on_destruct )
      : to_destruct( true )
      , resource( resourced )
      , destruct( on_destruct )
      {
      }

      /**
       * copy constructor.
       */
      result( result const& other )
      : to_destruct( other.to_destruct )
      , resource   ( other.resource    )
      , destruct   ( other.destruct    )
      {
         other.to_destruct = false;
      }

      /**
       * destruct resource, if required.
       */
      ~result()
      {
         if ( to_destruct )
            destruct( resource );
      }

      /**
       * return the resource, no destruction.
       */
      operator resource_type()
      {
         to_destruct = false;
         return resource;
      }

   private:
      /**
       * prevent assignment.
       */
      result& operator= ( result const& );

   private: // representation
      /**
       * destruct resource if true.
       */
      mutable bool to_destruct;

      /**
       * the resource.
       */
      resource_type resource;

      /**
       * the resource destructor.
       */
      destructor_type destruct;
   };

   /**
    * transfer resource with its ownership.
    */
   result transfer()
   {
      to_destruct = false;
      return result( resource, destruct );
   }

   /// @} Transfer

private:
   /**
    * destruct resource if true.
    */
   mutable bool to_destruct;

   /**
    * the resource.
    */
   resource_type resource;

   /**
    * the resource destructor.
    */
   destructor_type destruct;
};

} // namespace nonstd

#endif // _scoped_h_included

/*
 * end of file
 */
