/**
 * \file test-scoped.cpp test the nonstd::scoped<> class.
 *
 * Author: Martin Moene
 *
 * See also file scoped.h
 */

#include "scoped.h"     // for class scoped<>
#include <iostream>     // for std::cout

/**
 * standard deleter that prints evidence of the delete operation.
 */
struct printing_deleter
{
   /**
    * apply.
    */
   template< typename deleted_type >
   void operator()( deleted_type to_delete ) const
   {
      std::cout << "  deleter: deleting object" << std::endl;
      delete to_delete;
   }
};

/**
 * our type.
 */
struct type
{
   /**
    * print evidence of the construction.
    */
   type()
   {
      std::cout << "  type   : constructing object" << std::endl;
   }

   /**
    * print evidence of the destruction.
    */
   ~type()
   {
      std::cout << "  type   : destructing object" << std::endl;
   }
};

/**
 * the scoped type.
 */
typedef nonstd::scoped< type *, printing_deleter > scoped_type;

/**
 * transfer ownership.
 */
scoped_type::result example()
{
   scoped_type ptr( new type /* , printing_deleter() */ );

   return ptr.transfer();
}

/**
 * the test program.
 */
int main()
{
   /*
    * normal operation:
    */
   std::cout << "\nnormal scoped operation\n"
                " before scope" << std::endl;
   {
      scoped_type ptr( new type /* , printing_deleter() */ );
   }
   std::cout << " after scope: object must have been deleted" << std::endl;

   /*
    * transfer ownership:
    */
   type* ptr2;
   std::cout << "\ntransfer ownership\n"
                " before scope" << std::endl;
   {
      ptr2 = example();
   }
   std::cout << " after scope: object must NOT have been deleted" << std::endl;

   std::cout << " after scope: deleting object explicitly" << std::endl;
   delete ptr2;

   /*
    * transfer ownership, without using resource:
    */
   std::cout << "\ntransfer ownership, without using resource\n"
                " before scope" << std::endl;
   {
      example();
   }
   std::cout << " after scope: object must have been deleted" << std::endl;
}

/*
 * end of file
 */

